﻿using System;
using System.IO;
using System.Drawing;
using System.Drawing.Imaging;
using AdrianKousz.Util;

namespace AdrianKousz.GenieEngine
{
	public static class BitmapUtil
	{
		public static byte[] ReadRawBitmap(Stream input)
		{
			EnsureLittleEndian();

			var temp = input.ReadBytes(40);

			var dibHeaderLength = BitConverter.ToInt32(temp, 0);
			var bitDepth = BitConverter.ToInt32(temp, 14);
			var paletteLength = bitDepth > 8 ? 0 : (1 << bitDepth) * 4;
			var pixelArrayLength = BitConverter.ToInt32(temp, 20);
			var resultLength = dibHeaderLength + paletteLength + pixelArrayLength;

			if (pixelArrayLength == 0) {
				throw new NotImplementedException("Unspecified pixel array length");
			}

			var result = new byte[resultLength];
			temp.CopyTo(result, 0);

			input.Read(result, temp.Length, result.Length - temp.Length);

			return result;
		}

		public static Bitmap ToBitmap(this byte[] input)
		{
			EnsureLittleEndian();

			var bmpHeaderLength = 14;
			var dibHeaderLength = BitConverter.ToInt32(input, 0);
			var bitDepth = BitConverter.ToInt32(input, 14);
			var paletteLength = bitDepth > 8 ? 0 : (1 << bitDepth) * 4;
			var pixelArrayOffsetOutput = bmpHeaderLength + dibHeaderLength + paletteLength;
			var resultLength = input.Length + bmpHeaderLength;

			var array = new byte[resultLength];

			array[0] = 0x42;
			array[1] = 0x4d;
			BitConverter.GetBytes(resultLength).CopyTo(array, 2);
			BitConverter.GetBytes(pixelArrayOffsetOutput).CopyTo(array, 10);
			input.CopyTo(array, bmpHeaderLength);

			var result = new Bitmap(new MemoryStream(array));

			return result;
		}

		public static byte[] ToRawBitmap(this Bitmap input)
		{
			EnsureLittleEndian();

			var bmpHeaderLength = 14;
			var tempstream = new MemoryStream();
			input.Save(tempstream, ImageFormat.Bmp);
			tempstream.Dispose();
			var temp = tempstream.ToArray();

			var dibHeaderLength = BitConverter.ToInt32(temp, 0 + bmpHeaderLength);
			var bitDepth = BitConverter.ToInt32(temp, 14 + bmpHeaderLength);
			var paletteLength = bitDepth > 8 ? 0 : (1 << bitDepth) * 4;
			var pixelArrayOffsetInput = BitConverter.ToInt32(temp, 10);
			var pixelArrayLength = temp.Length - pixelArrayOffsetInput;
			var resultLength = dibHeaderLength + paletteLength + pixelArrayLength;

			var result = new byte[resultLength];
			Array.Copy(temp, bmpHeaderLength, result, 0, dibHeaderLength + paletteLength);
			Array.Copy(temp, pixelArrayOffsetInput, result, dibHeaderLength + paletteLength, pixelArrayLength);
			BitConverter.GetBytes(pixelArrayLength).CopyTo(result, 20);

			return result;
		}

		public static int GetWidth(this byte[] input)
		{
			EnsureLittleEndian();
			return BitConverter.ToInt32(input, 4);
		}

		public static int GetHeight(this byte[] input)
		{
			EnsureLittleEndian();
			return BitConverter.ToInt32(input, 8);
		}

		private static void EnsureLittleEndian()
		{
			if (!BitConverter.IsLittleEndian) {
				throw new NotSupportedException("Big-endian");
			}
		}
	}
}

