﻿using System.IO;
using System.IO.Compression;
using AdrianKousz.Util;

namespace AdrianKousz.GenieEngine
{
	public static class GenieFile
	{
		#region Serialization

		public static byte[] Serialize<T>(T value)
		{
			byte[] result;
			using (var stream = new MemoryStream())
			{
				Serialize(value, stream);
				result = stream.ToArray();
			}
			return result;
		}

		public static void Serialize<T>(T value, Stream output)
		{
			var scnValue = value as Scenario;
			if (scnValue != null) {
				var scnwriter = ScnSerializerWriter.CreateDefault();
				using (var writer = GetWriter(output))
				{
					scnwriter.Write(writer, scnValue);
					return;
				}
			}

			var cpnValue = value as Campaign;
			if (cpnValue != null) {
				var cpnwriter = CpnSerializerWriter.CreateDefault();
				using (var writer = GetWriter(output))
				{
					cpnwriter.Write(writer, cpnValue);
					return;
				}
			}

			throw new System.ArgumentException("Unsupported Type");
		}

		#endregion

		#region Deserialization

		public static T Deserialize<T>(byte[] array)
		{
			return (T)Deserialize(array, typeof(T));
		}

		public static T Deserialize<T>(Stream input)
		{
			return (T)Deserialize(input, typeof(T));
		}

		private static object Deserialize(byte[] array, System.Type type)
		{
			using (var stream = new MemoryStream(array))
			{
				return Deserialize(stream, type);
			}
		}

		private static object Deserialize(Stream input, System.Type type)
		{
			if (type == typeof(Scenario)) {
				var scnreader = ScnSerializerReader.CreateDefault();
				using (var reader = GetReader(input))
				{
					return scnreader.ReadScenario(reader);
				}
			}

			if (type == typeof(Campaign)) {
				var cpnreader = CpnSerializerReader.CreateDefault();
				using (var reader = GetReader(input))
				{
					return cpnreader.ReadCampaign(reader);
				}
			}

			throw new System.ArgumentException("Unsupported Type");
		}

		#endregion

		#region Low Level Methods

		public static ExtendedBinaryReader GetReader(Stream stream)
		{
			return new ExtendedBinaryReader(new NonDisposingStreamWrapper(stream), Scenario.Encoding);
		}

		public static ExtendedBinaryWriter GetWriter(Stream stream)
		{
			return new ExtendedBinaryWriter(new NonDisposingStreamWrapper(stream), Scenario.Encoding);
		}

		public static void ScenarioDecompress(Stream input, Stream output)
		{
			CopyHeader(input, output);

			using (var comp = new DeflateStream(input, CompressionMode.Decompress))
			{
				comp.CopyTo(output);
			}
		}

		public static void ScenarioCompress(Stream input, Stream output)
		{
			CopyHeader(input, output);

			using (var comp = new DeflateStream(output, CompressionMode.Compress))
			{
				input.CopyTo(comp);
			}
		}

		private static void CopyHeader(Stream input, Stream output)
		{
			var buffer = input.ReadBytes(8);
			output.Write(buffer);

			var headerLength = buffer[4]
				| buffer[5] << 8
				| buffer[6] << 16
				| buffer[7] << 24
			;

			buffer = input.ReadBytes(headerLength);
			output.Write(buffer);
		}

		#endregion
	}
}

