﻿using System.IO;
using System.IO.Compression;
using AdrianKousz.Util;

namespace AdrianKousz.GenieEngine
{
	internal class ScnSerializerWriter
	{
		public static ScnSerializerWriter CreateDefault()
		{
			return new ScnSerializerWriter();
		}

		private ScnSerializerWriter()
		{
		}

		public void Write(ExtendedBinaryWriter writer, Scenario value)
		{
			EnsureValid(value);

			var Version = value.VersionNumber;
			var UncompressedHeaderLength = 20 + writer.GetZByteCount(value.InstructionsUncompressed);

			// Uncompressed header

			writer.WriteStringRaw(value.VersionString);
			writer.Write((System.Int32)UncompressedHeaderLength);
			writer.Write(value.Unknown1);
			writer.Write((System.Int32)DateTimes.ToUnixTime(value.Timestamp));
			writer.WriteZStringInt32(value.InstructionsUncompressed);
			writer.Write(value.Unknown2);
			writer.Write(value.PlayerCount);

			// Compression starts

			var originalWriter = writer;
			var compressedPart = new MemoryStream();
			var comp = new DeflateStream(compressedPart, CompressionMode.Compress);
			writer = new ExtendedBinaryWriter(comp, originalWriter.Encoding);

			writer.Write(value.NextId);
			writer.Write(value.VersionNumber);

			// PlayerSettings 1

			value.PlayerSettings.ForEach(x => {
				writer.WriteZString(x.Name, 256);
			});

			if (Version >= 1.18f) {
				value.PlayerSettings.ForEach(x => {
					writer.Write(x.NameStringID);
				});
			}

			value.PlayerSettings.ForEach(x => {
				writer.WriteBooleanInt32(x.Active);
				writer.WriteBooleanInt32(x.Human);
				writer.Write(x.Civ);
				writer.Write(x.UnknownInfo);
			});

			writer.Write(value.Unknown3);
			writer.Write(value.Unknown4);
			writer.Write(value.Unknown5);

			writer.WriteStringInt16(value.Filename);

			// Messages

			if (Version >= 1.18f) {
				writer.Write(value.Messages.InstructionsStringID);
				writer.Write(value.Messages.HintsStringID);
				writer.Write(value.Messages.VictoryStringID);
				writer.Write(value.Messages.LossStringID);
				writer.Write(value.Messages.HistoryStringID);
			}
			if (Version >= 1.22f) {
				writer.Write(value.Messages.ScoutsStringID);
			}

			writer.WriteZStringInt16(value.Messages.Instructions);
			writer.WriteZStringInt16(value.Messages.Hints);
			writer.WriteZStringInt16(value.Messages.Victory);
			writer.WriteZStringInt16(value.Messages.Loss);
			writer.WriteZStringInt16(value.Messages.History);
			if (Version >= 1.22f) {
				writer.WriteZStringInt16(value.Messages.Scouts);
			}

			// Cinematics

			writer.WriteStringInt16(value.Cinematics.FilenamePregame);
			writer.WriteStringInt16(value.Cinematics.FilenameVictory);
			writer.WriteStringInt16(value.Cinematics.FilenameLoss);
			writer.WriteStringInt16(value.Cinematics.FilenameBitmap);

			var hasBitmap = value.Cinematics.RawBitmap.Length > 0;

			writer.WriteBooleanInt32(hasBitmap);
			writer.Write(value.Cinematics.BitmapWidth);
			writer.Write(value.Cinematics.BitmapHeight);
			writer.Write(value.Cinematics.BitmapUnknown);

			if (hasBitmap)
				writer.Write(value.Cinematics.RawBitmap);

			// PlayerSettings 2

			value.PlayerSettings.ForEach(x => {
				writer.WriteStringInt16(x.FilenameAI);
				writer.WriteStringInt16(x.FilenameCTY);
				writer.WriteStringInt16(x.FilenamePER);
			});

			value.PlayerSettings.ForEach(x => {
				writer.Write((System.Int32)writer.GetByteCount(x.ScriptAI));
				writer.Write((System.Int32)writer.GetByteCount(x.ScriptCTY));
				writer.Write((System.Int32)writer.GetByteCount(x.ScriptPER));
				writer.WriteStringRaw(x.ScriptAI);
				writer.WriteStringRaw(x.ScriptCTY);
				writer.WriteStringRaw(x.ScriptPER);
			});

			if (Version >= 1.18f) {
				value.PlayerSettings.ForEach(x => {
					writer.Write(x.AIType);
				});
			}

			WriteSeparator(writer, "Resources");

			// Resources

			value.PlayerSettings.ForEach(x => {
				writer.Write(x.ResourceGold);
				writer.Write(x.ResourceWood);
				writer.Write(x.ResourceFood);
				writer.Write(x.ResourceStone);
				if (Version >= 1.18f) {
					writer.Write(x.ResourceOre);
					if (Version < 1.3f) {
						writer.Write(x.UnknownResource);
					}
				}
			});

			WriteSeparator(writer, "Victory");

			// Diplomacy

			writer.WriteBooleanInt32(value.GlobalVictory.RequireConquest);
			writer.Write(value.GlobalVictory.Ruins);
			writer.Write(value.GlobalVictory.Artifacts);
			writer.Write(value.GlobalVictory.Discovery);
			writer.Write(value.GlobalVictory.PercentExplored);
			writer.Write(value.GlobalVictory.Unknown);
			writer.WriteBooleanInt32(value.GlobalVictory.RequireAllCustom);
			writer.Write(value.GlobalVictory.Mode);
			writer.Write(value.GlobalVictory.Score);
			writer.Write(value.GlobalVictory.Time);

			// PlayerSettings 3

			value.PlayerSettings.ForEach(x => {
				x.Diplomacy.ForEach(y => {
					writer.Write(y);
				});
			});

			writer.Write(value.RawIndividualVictory);

			WriteSeparator(writer, "Player Environment");

			value.PlayerSettings.ForEach(x => {
				writer.WriteBooleanInt32(x.AlliedVictory);
			});

			writer.Write(value.RawDisables);

			value.PlayerSettings.ForEach(x => {
				writer.Write(x.StartingAge);
			});

			WriteSeparator(writer, "Map");

			// Map

			if (Version >= 1.18f) {
				writer.Write(value.Map.CameraX);
				writer.Write(value.Map.CameraY);
			}
			if (Version >= 1.22f) {
				writer.Write(value.Map.GeneratorId);
			}
			writer.Write(value.Map.SizeX);
			writer.Write(value.Map.SizeY);

			value.Map.LinearTiles.ForEach(x => {
				writer.Write(x.Id);
				writer.Write(x.Elevation);
				writer.Write(x.Unknown);
			});

			// ResourceCopies

			writer.Write((System.Int32)value.Units.Count);

			value.ResourceCopies.ForEach(x => {
				writer.Write(x.ResourceFood);
				writer.Write(x.ResourceWood);
				writer.Write(x.ResourceGold);
				writer.Write(x.ResourceStone);
				if (Version >= 1.18f) {
					writer.Write(x.ResourceOre);
					if (Version < 1.3f) {
						writer.Write(x.UnknownResource);
					}
				}
				if (Version >= 1.22f) {
					writer.Write(x.PopulationLimit);
				}
			});

			// Units

			value.Units.ForEach(x => {
				writer.Write((System.Int32)x.Count);
				x.ForEach(y => {
					writer.Write(y.PosX);
					writer.Write(y.PosY);
					writer.Write(y.Unknown1);
					writer.Write(y.Id);
					writer.Write(y.UnitId);
					writer.Write(y.Unknown2);
					writer.Write(y.Rotation);
					if (Version >= 1.18f) {
						writer.Write(y.InitialFrame);
						writer.Write(y.GarrisonnedInId);
					}
				});
			});

			// PlayerSettings2

			writer.Write((System.Int32)(value.PlayerSettings2.Count + 1));

			value.PlayerSettings2.ForEach(x => {
				writer.WriteZStringInt16(x.Name);
				writer.Write(x.CameraX);
				writer.Write(x.CameraY);
				writer.Write(x.UnknownX);
				writer.Write(x.UnknownY);
				writer.Write(x.AlliedVictory);
				writer.Write((System.Int16)x.Stances.Length);
				x.Stances.ForEach(y => { writer.Write(y.Stance1); });
				x.Stances.ForEach(y => { writer.Write(y.Stance2); });
				if (Version >= 1.18f) {
					writer.Write(x.Color);
				}

				writer.Write(x.Unknown1);
				writer.Write(x.Unknown2);

				writer.Write(x.Unknown);
			});

			// Triggers

			writer.Write(value.RawRemaining);

			// End of writing

			writer.Dispose();
			comp.Dispose();
			compressedPart.Dispose();
			originalWriter.Write(compressedPart.ToArray());
		}

		public void WriteSeparator(ExtendedBinaryWriter writer, string name)
		{
			writer.Write(Scenario.Separator);
		}

		private void EnsureValid(Scenario value)
		{
			if (value.PlayerSettings.Count != Scenario.NumPlayers)
				throw new System.InvalidOperationException("PlayerSettings.Count != 16");
			if (value.Units.Count - 1 != value.ResourceCopies.Count)
				throw new System.InvalidOperationException("Conflicting unit and resource copy count");
		}
	}
}

